#BEGIN COPYRIGHT
#*************************************************************************
#
# Licensed Materials - Property of IBM
# 5725-C94, 5725-C95, 5725-C96
# (C) Copyright IBM Corporation 2011, 2012. All Rights Reserved.
# US Government Users Restricted Rights- Use, duplication or disclosure
# restricted by GSA ADP Schedule Contract with IBM Corp.
#
#*************************************************************************
#END COPYRIGHT

import codecs, os, os.path, re, time, shutil, string, sys
import AdminUtilities
from java.lang import System
from java.util.zip import ZipFile, ZipEntry
from java.util.jar import Manifest
from javax.management import ObjectName
from com.ibm.wbicore.migration.common import ProfileHelper

__file__ = "upgradeProfile.py"

execfile(os.path.join(System.getProperty("was.install.root"), "BPM/base/profile/actions/scripts/updateServerPolicyForBPM.py"))

#------------------------------------------------------------------------------
# Applications handled by this script - map application name prefix to EAR file
#------------------------------------------------------------------------------
apps = {}
apps.update( {"IBM_BPM_PerformanceDW": "BPM/Lombardi/performance-data-warehouse/applications/Twperfsvr.ear" } )
apps.update( {"IBM_BPM_Repository": "BPM/Lombardi/process-center/applications/Repository.ear" } )
apps.update( {"IBM_BPM_Help": "BPM/Lombardi/process-server/applications/BPMHelp.ear" } )
apps.update( {"IBM_BPM_Portal": "BPM/Lombardi/process-server/applications/Portal.ear" } )
apps.update( {"IBM_BPM_ProcessAdmin": "BPM/Lombardi/process-server/applications/ProcessAdmin.ear" } )
apps.update( {"IBM_BPM_Teamworks": "BPM/Lombardi/process-server/applications/Teamworks.ear" } )
apps.update( {"IBM_BPM_WebAPI": "BPM/Lombardi/process-server/applications/WebAPI.ear" } )
apps.update( {"IBM_BPM_Process_Portal": None } ) # ProcessPortal.ear has been removed in BPM 8.0.1
apps.update( {"IBM_BPM_Process_Portal_Notification": "BPM/Lombardi/process-server/applications/ProcessPortalNotification.ear" } )

#------------------------------------------------------------------------------
# log either to profile upgrade framework or to console, depending on context
#------------------------------------------------------------------------------
def log(level, method, message):
  if globals().has_key("framework"):
    framework.log(level, "upgradeBPMProfile.py", method, message)
  else:
    print message
#enddef

#------------------------------------------------------------------------------
# get the profile's config temp directory
#------------------------------------------------------------------------------
def getRepositoryTemp(suffix = ""):
  tempRoot= System.getProperty("was.repository.temp")
  if tempRoot:
    tempRoot= os.path.join(tempRoot, suffix) # need a normalized path
  else:
    # was.repository.temp is not available in connected mode where admin tasks run on the dmgr and cannot access the local file system
    tempRoot = os.path.join(System.getProperty("user.install.root"), "config/temp", suffix)
  #endif
  if suffix:
    pass # log("debug", "getRepositoryTemp", "  Returning temp directory " + tempRoot + " for suffix = " + suffix + ".")
  if not os.path.exists(tempRoot):
    os.makedirs(tempRoot)
  return tempRoot
#enddef

#------------------------------------------------------------------------------
# Get the list of nodes which are to be updated by the current invocation
#------------------------------------------------------------------------------
def getNodeNames():
  nodeNames = [ System.getProperty("local.node") ]
  managedNodes = System.getProperty("bpm.upgrade.managed.nodes")
  if managedNodes:
    nodeNames.extend(managedNodes[1:-1].split())
  #endif
  return nodeNames
#enddef

#------------------------------------------------------------------------------
# Get the containment path for a given scope
#------------------------------------------------------------------------------
def getScopePath(cellName, nodeName, serverName, clusterName):
  if clusterName:
    configPath = "/Cell:" + cellName + "/ServerCluster:" + clusterName + "/"
  elif serverName:
    configPath = "/Cell:" + cellName + "/Node:" + nodeName + "/Server:" + serverName + "/"
  elif nodeName:
    configPath = "/Cell:" + cellName + "/Node:" + nodeName + "/"
  else:
    configPath = "/Cell:" + cellName + "/"
  #endif
  return configPath
#enddef

#------------------------------------------------------------------------------
# Get the relative config path for a given scope
#------------------------------------------------------------------------------
def getConfigPath(cellName, nodeName, serverName, clusterName):
  if clusterName:
    configPath = "cells/" + cellName + "/clusters/" + clusterName + "/"
  elif serverName:
    configPath = "cells/" + cellName + "/nodes/" + nodeName + "/servers/" + serverName + "/"
  elif nodeName:
    configPath = "cells/" + cellName + "/nodes/" + nodeName + "/"
  else:
    configPath = "cells/" + cellName + "/"
  return configPath
#enddef

#------------------------------------------------------------------------------
# Get the relative config paths for a given scope
# For clusters, the templates to create new members will be at cluster scope
# but the runtime uses the config files at the individual members' scope's,
# so we need to update both
#------------------------------------------------------------------------------
def getConfigPaths(cellName, nodeName, serverName, clusterName, suffix):
  configPaths = [ getConfigPath(cellName, nodeName, serverName, clusterName) + suffix ]
  if clusterName:
    cluster = AdminConfig.getid(getScopePath(cellName, None, None, clusterName))
    members = AdminUtilities.convertToList(AdminConfig.showAttribute(cluster, "members"))
    for member in members:
      memberNode = AdminConfig.showAttribute(member, "nodeName")
      memberServer = AdminConfig.showAttribute(member, "memberName")
      configPaths += [ getConfigPath(cellName, memberNode, memberServer, None) + suffix ]
    #endfor
  #endif
  return configPaths
#enddef

#------------------------------------------------------------------------------
# Get all deployed applications with a given name prefix in a given scope
#------------------------------------------------------------------------------
def getAppNames(scope, appPrefix):
  appNames = [appName for appName in AdminUtilities.convertToList(AdminApp.list(scope)) if appName.startswith(appPrefix)]
  # prefix "IBM_BPM_Process_Portal" will have matched "IBM_BPM_Process_Portal_Notification", too. Need to remove those now.
  for appName in appNames[:]:
    # the candidate application's prefix is the longest matching prefix
    candPrefix = max([candPrefix for candPrefix in apps.keys() if appName.startswith(candPrefix)])
    if candPrefix != appPrefix:
      appNames.remove(appName)
  log("info", "getAppNames", "Applications with prefix " + appPrefix + ": " + str(appNames))
  return appNames
#enddef

#------------------------------------------------------------------------------
# Get the key / value pairs for a given task
#------------------------------------------------------------------------------
def getTaskInfo(uri, task, lines):
  options = []
  taskInfo = AdminUtilities.convertToList(AdminApp.view(uri, [task]))
  # the last <lines> lines contain the task fields (key / value pairs), e.g.:
  #   Web module:  BPM Help
  #   URI:  BPMHelp.war,WEB-INF/web.xml
  #   Virtual host:  null
  # caution: the key may be translated!
  pattern = re.compile("[^:]*:\s*(.*)")
  for line in taskInfo[-lines:]:
    matched = pattern.match(line)
    if matched:
      options += [matched.group(1)]
  #endfor
  return options
#enddef

#------------------------------------------------------------------------------
# Get all deployed applications with a given name prefix in a given scope
#------------------------------------------------------------------------------
def getAppUpdateOptions(appName, appEAR):
  options = []
  moduleURIs = AdminUtilities.convertToList(AdminApp.listModules(appName))
  log("info", "getAppUpdateOptions", "  Modules for " + appName + ": " + str(moduleURIs))
  if appName.startswith("IBM_BPM_Teamworks"):
    options += ["-usedefaultbindings"]
    tmoptions = []
    webviewerURI = appName + "#webviewer.war+WEB-INF/web.xml"
    if not webviewerURI in moduleURIs:
      # module bpmjaxws.war was added in 7.5.1, copy module to server mapping (tmoptions[2]) from teamworks.war
      tmoptions1 = getTaskInfo(appName + "#teamworks.war+WEB-INF/web.xml", "-MapModulesToServers", 3)
      tmoptions1[0] = ".*"
      tmoptions1[1] = "webviewer.war,WEB-INF/web.xml"
      tmoptions += [tmoptions1]
      # and copy virtual host mapping (vhoptions[2]) from teamworks.war
      vhoptions1 = getTaskInfo(appName + "#teamworks.war+WEB-INF/web.xml", "-MapWebModToVH", 3)
      vhoptions1[0] = ".*"
      vhoptions1[1] = "webviewer.war,WEB-INF/web.xml"
      options += ["-MapWebModToVH", [vhoptions1]]
    #endif
    BPMSearchStartupBeanURI = appName + "#BPMSearchStartupBean.jar,META-INF/ejb-jar.xml"
    if not BPMSearchStartupBeanURI in moduleURIs:
      # modules BPMSearchStartupBean.jar and SharingNotificationBean.jar were added in 8.0.0, copy module to server mapping (tmoptions[2]) from twcore-ejb.jar
      tmoptions2 = getTaskInfo(appName + "#twcore-ejb.jar+META-INF/ejb-jar.xml", "-MapModulesToServers", 3)
      tmoptions2[0] = ".*"
      tmoptions2[1] = "BPMSearchStartupBean.jar,META-INF/ejb-jar.xml"
      tmoptions += [tmoptions2]
      tmoptions3 = [tmoptions2[0], "SharingNotificationBean.jar,META-INF/ejb-jar.xml", tmoptions2[2]]
      tmoptions += [tmoptions3]
    #endif
    ServerCommandStartupURI = appName + "#ServerCommandStartupBean.jar+META-INF/ejb-jar.xml"
    if not ServerCommandStartupURI in moduleURIs:
      # modules ServerCommandStartupBean.jar and bpmasset.war were added in 8.0.1, copy module to server mapping (tmoptions[2]) from twcore-ejb.jar
      tmoptions4 = getTaskInfo(appName + "#twcore-ejb.jar+META-INF/ejb-jar.xml", "-MapModulesToServers", 3)
      tmoptions4[0] = ".*"
      tmoptions4[1] = "ServerCommandStartupBean.jar,META-INF/ejb-jar.xml"
      tmoptions += [tmoptions4]
      tmoptions5 = [tmoptions4[0], "bpmasset.war,WEB-INF/web.xml", tmoptions4[2]]
      tmoptions += [tmoptions5]
      # and copy virtual host mapping (vhoptions[2]) from teamworks.war
      vhoptions2 = getTaskInfo(appName + "#teamworks.war+WEB-INF/web.xml", "-MapWebModToVH", 3)
      vhoptions2[0] = ".*"
      vhoptions2[1] = "bpmasset.war,WEB-INF/web.xml"
      options += ["-MapWebModToVH", [vhoptions2]]
    #endif
    if tmoptions:
      options += ["-MapModulesToServers", tmoptions]
  #endif
  if appName.startswith("IBM_BPM_WebAPI"):
    tmoptions = []
    bpmjaxwsURI = appName + "#bpmjaxws.war+WEB-INF/web.xml"
    if not bpmjaxwsURI in moduleURIs:
      # modules bpmjaxws.war and bpmjaxwsejb.jar were added in 7.5.1, copy module to server mapping (tmoptions[2]) from webapi.war
      tmoptions1 = getTaskInfo(appName + "#webapi.war+WEB-INF/web.xml", "-MapModulesToServers", 3)
      tmoptions1[0] = ".*"
      tmoptions1[1] = "bpmjaxws.war,WEB-INF/web.xml"
      tmoptions += [tmoptions1]
      tmoptions2 = [tmoptions1[0], "bpmjaxwsejb.jar,META-INF/ejb-jar.xml", tmoptions1[2]]
      tmoptions += [tmoptions2]
      # and copy virtual host mapping (vhoptions[2]) from webapi.war
      vhoptions1 = getTaskInfo(appName + "#webapi.war+WEB-INF/web.xml", "-MapWebModToVH", 3)
      vhoptions1[0] = ".*"
      vhoptions1[1] = "bpmjaxws.war,WEB-INF/web.xml"
      options += ["-MapWebModToVH", [vhoptions1]]
      # and set 'Everyone' authentication for new role "BPMWSAPIUser"
      options += ["-MapRolesToUsers", [["BPMWSAPIUser", "Yes", "", "", "", "", ""]]]
    #endif
    integration_clients_ejbURI = appName + "#integration-clients-ejb.jar,META-INF/ejb-jar.xml"
    if not integration_clients_ejbURI in moduleURIs:
      # module integration-clients-ejb.jar was added in 8.0.0, copy module to server mapping (tmoptions[2]) from webapi.war
      tmoptions3 = getTaskInfo(appName + "#webapi.war+WEB-INF/web.xml", "-MapModulesToServers", 3)
      tmoptions3[0] = ".*"
      tmoptions3[1] = "integration-clients-ejb.jar,META-INF/ejb-jar.xml"
      tmoptions += [tmoptions3]
    #endif
    if tmoptions:
      options += ["-MapModulesToServers", tmoptions]
  #endif
  if appName.startswith("IBM_BPM_Portal"):
    tmoptions = []
    vhoptions = []
    processURI = appName + "#portal.war+WEB-INF/web.xml"
    if processURI in moduleURIs:
      # module portal.war was replaced by process-portal.war and process-portal-support.war in 8.0.1, copy module to server mapping (tmoptions[2]) from portal.war
      tmoptions1 = getTaskInfo(appName + "#portal.war+WEB-INF/web.xml", "-MapModulesToServers", 3)
      tmoptions1[0] = ".*"
      tmoptions1[1] = "process-portal.war,WEB-INF/web.xml"
      tmoptions += [tmoptions1]
      tmoptions2 = tmoptions1
      tmoptions2 = [tmoptions1[0], "process-portal-support.war,WEB-INF/web.xml", tmoptions1[2]]
      tmoptions += [tmoptions2]
      # and copy virtual host mapping (vhoptions[2]) from portal.war
      vhoptions1 = getTaskInfo(appName + "#portal.war+WEB-INF/web.xml", "-MapWebModToVH", 3)
      vhoptions1[0] = ".*"
      vhoptions1[1] = "process-portal.war,WEB-INF/web.xml"
      vhoptions += [vhoptions1]
      vhoptions2 = [vhoptions1[0], "process-portal-support.war,WEB-INF/web.xml", vhoptions1[2]]
      vhoptions += [vhoptions2]
    #endif
    if tmoptions:
      options += ["-MapModulesToServers", tmoptions]
    if vhoptions:
      options += ["-MapWebModToVH", vhoptions]
  #endif
  if appName.startswith("IBM_BPM_PerformanceDW"):
    # to bind new EJBs to the default JNDI names
    options += ["-usedefaultbindings"]
  #endif
  if appName.startswith("IBM_BPM_Help"):
    # to assign "default_host" as virtual host
    options += ["-usedefaultbindings"]
  #endif
  return options
#enddef

#------------------------------------------------------------------------------
# Update the given infrastructure application at the given scope.
#------------------------------------------------------------------------------
def updateApplication(appName, nodeName, serverName, clusterName):
  # will return sth. like 'Application Build ID:  District9-Release-82-50fa14e94bb1636ac1d89c203cec33838b10be22'
  deployedVersion = AdminApp.view(appName, ["-buildVersion"])
  # the word after the first colon in the string is the actual version identifier
  deployedVersion = deployedVersion.split(":", 1)[1].lstrip()
  log("info", "updateApplication", "  Deployed version: " + deployedVersion + ".")

  # there may be several matching prefixes, use the longest one (through 'max()')
  candPrefixes = [appPrefix for appPrefix in apps.keys() if appName.startswith(appPrefix)]
  if candPrefixes:
    # appEAR may be "None" which indicates that the application no longer exists and hence should be uninstalled
    appEAR = apps.get(max(candPrefixes))
    if appEAR:
      appEAR = os.path.join(System.getProperty("was.install.root"), appEAR)
    if not appEAR:
      log("info", "updateApplication", "  Uninstalling " + appName)
      AdminApp.uninstall(appName)
    elif not os.path.isfile(appEAR):
      log("warning", "updateApplication", "Warning: File " + appEAR + " not found, skipping check.")
    else:
      log("info", "updateApplication", "  Installed file: " + appEAR + ".")
      zf = ZipFile(appEAR)
      ze = ZipEntry('META-INF/MANIFEST.MF')
      mf = Manifest(zf.getInputStream(ze))
      installedVersion = mf.getMainAttributes().getValue("Implementation-Version")
      log("info", "updateApplication", "  Installed version: " + installedVersion + ".")
      if installedVersion != deployedVersion:
        options = ["-update", "-appname", appName, "-nodeployejb"]
        options += getAppUpdateOptions(appName, appEAR)
        log("info", "updateApplication", "  Updating " + appName + " with options " + str(options))
        AdminApp.install(appEAR, options)
      #endif
    #endif
  #endif
#enddef

#------------------------------------------------------------------------------
# Create the ME destination location variable at the specified scope if
# it doesn't exist yet and if the destination location is unique
#------------------------------------------------------------------------------
def createMEDestinationVariable(varName, busName, nodeName, serverName, clusterName):
  if clusterName:
    varScope = "Cell=" + cellName + ",Cluster=" + clusterName
  else:
    varScope = "Cell=" + cellName + ",Node=" + nodeName + ",Server=" + serverName
  varValue = AdminTask.showVariables(["-variableName", varName, "-scope", varScope])
  if varValue:
    log("info", "createMEDestinationVariable", varName + " already set to " + varValue + " at scope " + varScope)
  else:
    busMembers = AdminUtilities.convertToList(AdminTask.listSIBusMembers(['-bus', busName]))
    if len(busMembers) != 1:
      log("info", "createMEDestinationVariable", "Cannot create " + varName + " because " + busName + " doesn't have exactly one member: " + str(busMembers))
      log("info", "createMEDestinationVariable", "Create " + varName + " manually on scope " + varScope + ", then re-run BPMProfileUpgrade.ant.")
      raise UserWarning, "Create " + varName + " manually on scope " + varScope + ", then re-run BPMProfileUpgrade.ant."
    else:
      (meNode, meServer, meCluster) = [AdminConfig.showAttribute(busMembers[0], attr) for attr in ["node", "server", "cluster"]]
      if meCluster:
        varValue = "WebSphere:cluster=" + meCluster
      else:
        varValue = "WebSphere:node=" + meNode + ",server=" + meServer
      AdminTask.setVariable(["-variableName", varName, "-variableValue", varValue, "-scope", varScope])
      log("info", "createMEDestinationVariable", varName + " has been set to " + varValue + " at scope " + varScope)
    #endif
  #endif
  return varValue
#enddef

#------------------------------------------------------------------------------
# Update the configuration files for a given infrastructure application
# at a given scope
#------------------------------------------------------------------------------
def updateConfigFiles(appName, nodeName, serverName, clusterName):
  if appName.startswith("IBM_BPM_PerformanceDW"):
    updatePerformanceDW(nodeName, serverName, clusterName)
  elif appName.startswith("IBM_BPM_Teamworks"):
    configPath = getConfigPath(cellName, nodeName, serverName, clusterName)
    # we need a file as the tracer here, a plain directory will not do
    isProcessCenterServer = AdminConfig.existsDocument(configPath + "process-center/config/system/00Static.xml")
    updateProcessServer(nodeName, serverName, clusterName, isProcessCenterServer)
  #endif
#enddef

#------------------------------------------------------------------------------
# Return the given file's contents, assuming a UTF-8 encoding
#------------------------------------------------------------------------------
def readFileUTF8(file):
  file = codecs.open(file, "r", "utf-8")
  content = file.read()
  file.close()
  return content
#enddef

#------------------------------------------------------------------------------
# Replace a regex in a given file by some given string
#------------------------------------------------------------------------------
def replaceInFile(tempFile, pattern, replacement):
  content = readFileUTF8(tempFile)
  log("debug", "replaceInFile", "  Replacing " + pattern + " by " + replacement + " in " + tempFile + ".")
  (content, updated) = re.subn(pattern, replacement, content)
  if updated:
    file = codecs.open(tempFile, "w", "utf-8")
    file.write(content)
    file.close()
  #endif
  return updated
#enddef

#------------------------------------------------------------------------------
# Update the configuration files for a given infrastructure application
# at a given scope
# configPath - the configuration path as returned by getConfigPath()
# filePath - config file path relative to configPath
# pattern - regular expression specifying what to replace
# replacement - replacement for pattern
#------------------------------------------------------------------------------
def updateConfigFile(configPath, filePath, pattern, replacement):

  configFile = configPath + filePath

  if not AdminConfig.existsDocument(configFile):
    log("warning", "updateConfigFile", "  Configuration file " + configFile + " does not exist, skipping")
    return
  #endif

  tempDir = getRepositoryTemp(os.path.dirname(configFile))
  fileName = os.path.basename(configFile)
  tempFile = os.path.join(tempDir, fileName)
  log("info", "updateConfigFile", "  Extracting " + configFile)
  digest = AdminConfig.extract(configFile, tempFile)

  if not replaceInFile(tempFile, pattern, replacement):
    log("info", "updateConfigFile", "  Already up-to-date, skipping.")
  else:
    log("info", "updateConfigFile", "  Checking in " + configFile + "...")
    AdminConfig.checkin(configFile, tempFile, digest)
  #endif

  # clean up the temp file and the temp directories
  shutil.rmtree(getRepositoryTemp(re.split(r"/", configFile, 1)[0]))
#enddef

#------------------------------------------------------------------------------
# This function first extracts the current values of the given variables from
# the config XML file, then updates the config XML file from the corresponding
# template file, replacing the variables in there with their previous values.
#------------------------------------------------------------------------------
def copyTemplateFile(templatePath, configFile, variables):
  if not AdminConfig.existsDocument(configFile):
    log("warning", "copyTemplateFile", "  Configuration file " + configFile + " does not exist, skipping")
    return None
  #endif

  tempDir = getRepositoryTemp(os.path.dirname(configFile))
  fileName = os.path.basename(configFile)
  tempFile = os.path.join(tempDir, fileName)
  log("info", "copyTemplateFile", "  Extracting " + configFile)
  digest = AdminConfig.extract(configFile, tempFile)
  oldcontent = readFileUTF8(tempFile)

  replacements = {}
  for variableName in variables.keys():
    (fromDelim, toDelim) = variables[variableName]
    match = re.search(fromDelim + '(.*?)' + toDelim, oldcontent)
    if not match:
      replacements[variableName] = None
      log("info", "copyTemplateFile", "  Variable " + variableName + " not found in " + configFile)
    else:
      replacements[variableName] = match.group(1)
      log("info", "copyTemplateFile", "  Read " + variableName + " = '" + replacements[variableName] + "' from " + configFile)
    #endif
  #endfor

  templateFile = os.path.join(templatePath, fileName)
  if not os.path.exists(templateFile):
    # template for '50AppServer.xml' is '50AppServer.websphere.xml'
    templateFile = os.path.splitext(templateFile)[0] + ".websphere" + os.path.splitext(templateFile)[1]
  newcontent = readFileUTF8(templateFile)

  for variableName, replacement in replacements.items():
    if replacement != None:
      newcontent = re.sub(variableName, replacement, newcontent)
    else:
      newcontent = specialCaseReplacement(configFile, variableName, oldcontent, newcontent)
  #endfor

  if oldcontent == newcontent:
    log("info", "copyTemplateFile", "  No changes in " + configFile + ", skipping")
  else:
    log("info", "copyTemplateFile", "  Backing up " + configFile + " to " + backupDir)
    backupFile = os.path.join(backupDir, configFile)
    if not os.path.exists(os.path.dirname(backupFile)):
      os.makedirs(os.path.dirname(backupFile))
    shutil.copy(tempFile, backupFile)
    log("info", "copyTemplateFile", "  Checking in updated " + configFile + "...")
    file = codecs.open(tempFile, "w", "utf-8")
    file.write(newcontent)
    file.close()
    AdminConfig.checkin(configFile, tempFile, digest)
  #endif

  # clean up the temp file and the temp directories
  shutil.rmtree(getRepositoryTemp(re.split(r"/", configFile, 1)[0]))
#enddef

#------------------------------------------------------------------------------
# Extract the database type from a given 60Database.xml
#------------------------------------------------------------------------------
def getDatabaseType(databaseXML):
  file = open(databaseXML, "r")
  content = file.read()
  file.close()
  databaseType = re.sub(r'(?s).*<database-type>([^<]+)</database-type>.*', r'\1', content)
  log("info", "getDatabaseType", "  Database type of " + databaseXML + " is " + databaseType)
  return databaseType
#enddef

#------------------------------------------------------------------------------
# Finds 60Database.xml for a given DB type in a given list of files and checks it in
#------------------------------------------------------------------------------
def copy60DatabaseXML(args, dirname, names):
  # for the meaning of 'args' see os.path.walk() in update60DatabaseXML() below
  (databaseType, configFile, tempFile, digest) = args
  for name in names:
    if name == "60Database.xml":
      databaseXML = os.path.join(dirname, name)
      currentDBType = getDatabaseType(databaseXML)
      if currentDBType == databaseType:
        file = open(databaseXML, "r")
        content = file.read()
        file.close()
        file = open(tempFile, "w")
        file.write(content)
        file.close()
        log("info", "copy60DatabaseXML", "  Checking in " + configFile + "...")
        AdminConfig.checkin(configFile, tempFile, digest)
        return
      #endif
    #endif
  #endfor
#enddef

#------------------------------------------------------------------------------
# Update the 60Database.xml file on a given scope
#------------------------------------------------------------------------------
def update60DatabaseXML(configPath):

  filePath = "config/system/60Database.xml";

  tempDir = getRepositoryTemp(os.path.join(configPath, os.path.dirname(filePath)))
  fileName = os.path.basename(filePath)
  tempFile = os.path.join(tempDir, fileName)
  configFile = configPath + filePath

  log("info", "update60DatabaseXML", "  Extracting " + configFile)
  digest = AdminConfig.extract(configFile, tempFile)
  databaseType = getDatabaseType(tempFile)

  if os.path.basename(configPath) == "performance-data-warehouse":
    templatePath = os.path.join(System.getProperty("was.install.root"), "BPM/Lombardi/performance-data-warehouse/configTemplates/db")
  else:
    # caution - basename might as well be "process-center" here
    templatePath = os.path.join(System.getProperty("was.install.root"), "BPM/Lombardi/process-server/configTemplates/db")
  #endif

  # only one argument can be passed to the visitor, so we use a tuple that is unpacked in copy60DatabaseXML() above
  os.path.walk(templatePath, copy60DatabaseXML, (databaseType, configFile, tempFile, digest))

  # clean up the temp file and the temp directories
  shutil.rmtree(getRepositoryTemp(re.split(r"/", configPath, 1)[0]))
#enddef

#------------------------------------------------------------------------------
# Handles special cases when replacing variables in template XML files.
# For now, this will take care of migrating the BPM 7.5.0 user/id password
# entries that have been replaced by authentication data aliases in BPM 7.5.1
#------------------------------------------------------------------------------
def specialCaseReplacement(configFile, variableName, oldcontent, newcontent):
  def copySnippet(fromDelim, toDelim, oldcontent, newcontent):
    match = re.search(fromDelim + r'(.*?)' + toDelim, oldcontent)
    if match:
      newcontent = re.sub(r'(' + fromDelim + r').*?(' + toDelim + r')', r'\1' + match.group(1) + r'\2', newcontent)
    return newcontent
  #enddef
  if re.search(r"/process-(?:center|server)/.*/50AppServer.xml", configFile) and variableName == "~BPMJMS_AUTH_ALIAS~":
    # ~BPMJMS_AUTH_ALIAS~ from BPM 7.5.1 on
    newcontent = copySnippet(r'</cache>', '</common>', oldcontent, newcontent)
    log("info", "specialCaseReplacement", "  Performed custom replacement of " + variableName + " in " + configFile)
  elif re.search(r"/process-(?:center|server)/.*/99Local.xml", configFile) and variableName == "~REPOSITORY_SERVER_USER_AUTH_ALIAS~":
    newcontent = copySnippet(r'<server .*?<webservices ', '<timeout>', oldcontent, newcontent)
    log("info", "specialCaseReplacement", "  Performed custom replacement of 'BPMWebservice_Auth_Alias' in " + configFile)
    newcontent = copySnippet(r'<system-lane-users>', r'</system-lane-users>', oldcontent, newcontent)
    log("info", "specialCaseReplacement", "  Performed custom replacement of 'BPMAdmin_Auth_Alias' in " + configFile)
    newcontent = copySnippet(r'</repository-server-url>', r'<repository-server-interval>', oldcontent, newcontent)
    log("info", "specialCaseReplacement", "  Performed custom replacement of 'BPMAuthor_Auth_Alias' in " + configFile)
    log("info", "specialCaseReplacement", "  Performed custom replacement of " + variableName + " in " + configFile)
  else:
    log("warning", "specialCaseReplacement", "  Warning: Don't know how to handle " + variableName + " in " + configFile + ", skipping.")
  #endif
  return newcontent
#enddef

#------------------------------------------------------------------------------
# Update the process server configuration files at a given scope
#------------------------------------------------------------------------------
def updateProcessServer(nodeName, serverName, clusterName, isProcessCenterServer):
  if isProcessCenterServer:
    log("info", "updateProcessServer", "Updating Process Center configuration files")
    suffix = "process-center/"
  else:
    log("info", "updateProcessServer", "Updating Process Server configuration files")
    suffix = "process-server/"
  #endif
  variables00Static = {}
  variables00Static["~ENABLE_TS_COMM~"] = (r'<performance-server-communication>.*?<enabled>', r'</enabled>')
  variables00Static["~TWPERFSERVER_JNDIURL~"] = (r'<performance-server-communication>.*?<url>', r'</url>')
  variables00Static["~TWPERFSERVER_JNDI_SCOPE~"] = (r'<performance-server-communication>.*?<jndi-scope>', r'</jndi-scope>')
  variables50AppServer = {}
  variables50AppServer["~BPMJMS_AUTH_ALIAS~"] = (r'<jms-auth .*?<jms-user-auth-alias>', r'</jms-user-auth-alias>')
  variables50AppServer["~APPSERVER_JNDIURL~"] = (r'<server .*?<jndi-url>', r'</jndi-url>')
  variables50AppServer["~RUNTIME_LOGPATH~/~APPSERVER_SERVER~"] = (r'<log-watcher>.*?file-name="', r'/SystemOut.log"')
  variables50AppServer["~DEPLOY_DIR~"] = (r'<xml-catalog .*?<base-uri>', r'/webservices/</base-uri>')
  variables50AppServer["~APPSERVER_CELL~"] = (r'<server .*?<ear-location>installedApps/', r'/teamworks.ear</ear-location>')
  variables98Database = {}
  variables98Database["~DATABASEDRIVER~"] = (r'<database type="default">.*?<driver>', r'</driver>')
  variables98Database["~DATABASEURL~"] = (r'<database type="default">.*?<url>', r'</url>')
  variables98Database["~USERNAME~"] = (r'<database type="default">.*?<user>', r'</user>')
  variables98Database["~SCHEMA~"] = (r'<database type="default">.*?<schema>', r'</schema>')
  variables98Database["~DATABASE_ENCRYPTED_PASSWORD~"] = (r'<database type="default">.*?<encrypted-password>', r'</encrypted-password>')
  variables98Database["~DATABASE_NAME~"] = (r'<database type="default">.*?<db-name>', r'</db-name>')
  variables98Database["~STORAGE_GROUP~"] = (r'<database type="default">.*?<storage-group>', r'</storage-group>')
  variables99Local = {}
  variables99Local["~APPSERVER_URL~"] = (r'<authoring-environment .*?<repository-prefix>', r'/ProcessCenter</repository-prefix>')
  variables99Local["~APPSERVER_ENVNAME~"] = (r'<common .*?<environment-name>', r'</environment-name>')
  variables99Local["~TEMP_DIR~"] = (r'<common .*?<temporary-directory>', r'</temporary-directory>')
  variables99Local["~APPSERVER_SMTP_SERVER~"] = (r'<server .*?<smtp-server>', r'</smtp-server>')
  variables99Local["~REPOSITORY_SERVER_URL~"] = (r'<server .*?<repository-server-url>', r'/ProcessCenter</repository-server-url>')
  variables99Local["~REPOSITORY_SERVER_USER_AUTH_ALIAS~"] = (r'<server .*?<repository-server-user-auth-alias>', r'</repository-server-user-auth-alias>')
  variables99Local["~REPOSITORY_SERVER_HEARTBEAT_INTERVAL~"] = (r'<server .*?<repository-server-interval>', r'</repository-server-interval>')
  variables99Local["~SERVER_NAME~"] = (r'<server .*?<server-name>', r'</server-name>')
  variables99Local["~SERVER_DESCRIPTION~"] = (r'<server .*?<server-description>', r'</server-description>')
  variables99Local["~APPSERVER_HOST~"] = (r'<server .*?<server-host>', r'</server-host>')
  variables99Local["~APPSERVER_PORT~"] = (r'<server .*?<server-port>', r'</server-port>')
  variables99Local["~ENVIRONMENT_TYPE~"] = (r'<server .*?<environment-type>', r'</environment-type>')
  templatePath = os.path.join(System.getProperty("was.install.root"), "BPM/Lombardi/process-server/configTemplates/system")
  for configPath in getConfigPaths(cellName, nodeName, serverName, clusterName, suffix):
    copyTemplateFile(templatePath, configPath + "config/system/00Static.xml", variables00Static)
    copyTemplateFile(templatePath, configPath + "config/system/50AppServer.xml", variables50AppServer)
    update60DatabaseXML(configPath)
    copyTemplateFile(templatePath, configPath + "config/system/98Database.xml", variables98Database)
    copyTemplateFile(templatePath, configPath + "config/system/99Local.xml", variables99Local)
  #endfor
#enddef

#------------------------------------------------------------------------------
# Update the performance data warehouse configuration files at a given scope
#------------------------------------------------------------------------------
def updatePerformanceDW(nodeName, serverName, clusterName):
  log("info", "updatePerformanceDW", "Updating Performance Data Warehouse configuration files")
  variables00Static = {}
  variables00Static["~TWPERFSERVER_AUTO_SCHEMA_MGMT~"] = (r'<load-tracking-definition>.*?<automatic-schema-management>', r'</automatic-schema-management>')
  variables00Static["~TWPERFSERVER_USERNAME~"] = (r'<system name="TeamWorksProcessServer">.*?<view-user>', r'</view-user>')
  variables98Database = {}
  variables98Database["~TWPERFSERVER_DATABASEDRIVER~"] = (r'<database type="default">.*?<driver>', r'</driver>')
  variables98Database["~TWPERFSERVER_DATABASEURL~"] = (r'<database type="default">.*?<url>', r'</url>')
  variables98Database["~TWPERFSERVER_USERNAME~"] = (r'<database type="default">.*?<user>', r'</user>')
  variables98Database["~SCHEMA~"] = (r'<database type="default">.*?<schema>', r'</schema>')
  variables98Database["~DATABASE_ENCRYPTED_PASSWORD~"] = (r'<database type="default">.*?<encrypted-password>', r'</encrypted-password>')
  variables98Database["~DATABASE_NAME~"] = (r'<database type="default">.*?<db-name>', r'</db-name>')
  variables98Database["~STORAGE_GROUP~"] = (r'<database type="default">.*?<storage-group>', r'</storage-group>')
  templatePath = os.path.join(System.getProperty("was.install.root"), "BPM/Lombardi/performance-data-warehouse/configTemplates/system")
  for configPath in getConfigPaths(cellName, nodeName, serverName, clusterName, "performance-data-warehouse/"):
    copyTemplateFile(templatePath, configPath + "config/system/00Static.xml", variables00Static)
    update60DatabaseXML(configPath)
    copyTemplateFile(templatePath, configPath + "config/system/98Database.xml", variables98Database)
  #endfor
#enddef

#------------------------------------------------------------------------------
# This function implements the profile upgrade framework definition that accepts
# an appPrefix, nodeName, serverName, and clusterName).  It will be called by the
# framework function updateDeploymentTargetsOf.  Either the nodeName/serverName
# pair will be specified or the clusterName parameter will be specified.
# The others will be None.
#------------------------------------------------------------------------------
def updateConfiguration(appName, nodeName, serverName, clusterName):
  configPath = getScopePath(cellName, nodeName, serverName, clusterName)
  log("info", "updateConfiguration", "Processing " + appName + " configuration on " + configPath + "...")
  if not clusterName:
    server = AdminConfig.getid(configPath)
    if AdminConfig.showAttribute(server, "serverType") != "APPLICATION_SERVER":
      log("info", "updateConfiguration", "Deployment target " + configPath + " is not an application server, skipping.")
      return
    #endif
  #endif
  # caution: resources must be created before updating the application
  if appName.startswith("IBM_BPM_Teamworks") or appName.startswith("IBM_BPM_WebAPI") or appName.startswith("IBM_BPM_Process_Portal_Notification"):
    procSvrBusName = "PROCSVR." + cellName + ".Bus"
    createMEDestinationVariable("PROCSVR_ME_DESTINATION_LOCATION", procSvrBusName, nodeName, serverName, clusterName)
  #endif
  if appName.startswith("IBM_BPM_PerformanceDW"):
    perfDWBusName = "PERFDW." + cellName + ".Bus"
    createMEDestinationVariable("PERFDW_ME_DESTINATION_LOCATION", perfDWBusName, nodeName, serverName, clusterName)
  #endif
  updateApplication(appName, nodeName, serverName, clusterName)
  updateConfigFiles(appName, nodeName, serverName, clusterName)
  def notnull(val):
    if val: return val
    else: return ""
  #enddef
  params = [notnull(param) for param in ["-applicationName", appName, "-nodeName", nodeName, "-serverName", serverName, "-clusterName", clusterName]]
  AdminTask.upgradeBPMConfiguration(params)
  log("info", "updateConfiguration", "Finished processing " + appName + ".")
#enddef

#------------------------------------------------------------------------------
# This function changes the host name, was install root and profile root
# in the WLE config files for all servers under the current
#------------------------------------------------------------------------------
def changeHardware():
  oldHostName = System.getProperty("bpm.upgrade.previous.hostName");
  newHostName = System.getProperty("bpm.upgrade.current.hostName");

  log("info", "changeHardware", "old: "+ oldHostName + " new: " + newHostName)

  oldWasRoot = System.getProperty("bpm.upgrade.previous.was.root")
  newWasRoot = System.getProperty("was.install.root")

  log("info", "changeHardware", "old: " + oldWasRoot + " new: " + newWasRoot)

  oldProfileRoot = System.getProperty("bpm.upgrade.previous.profile.root")
  newProfileRoot = System.getProperty("user.install.root")

  log("info", "changeHardware", "old: " + oldProfileRoot + " new: " + newProfileRoot)

  nodeName = System.getProperty("local.node")
  servers = AdminUtilities.convertToList(AdminTask.listServers(["-serverType", "APPLICATION_SERVER", "-nodeName", nodeName]))

  for server in servers:
    serverName = AdminConfig.showAttribute(server, "name")
    configPath = getConfigPath(cellName, nodeName, serverName, None)

    log("info", "changeHardware", "Processing deployment target " + configPath + ".")

    suffix = "process-center/"
    document = configPath + suffix + "config/system/00Static.xml"
    if AdminConfig.existsDocument(document):
      configPath = configPath + suffix
      updateConfigFile(configPath, "config/system/99Sharepoint.xml",r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/system/99Local.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/system/80EventManager.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/HTTPEventListener.properties", r''+oldHostName+'',r''+newHostName+'')

      #update profile root
      updateConfigFile(configPath, "config/system/99Local.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')
      updateConfigFile(configPath, "config/log4j.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')

      #update install root
      updateConfigFile(configPath, "config/system/99Local.xml", r''+oldWasRoot+'',r''+newWasRoot+'')
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldWasRoot+'',r''+newWasRoot+'')
      updateConfigFile(configPath, "config/log4j.xml", r''+oldWasRoot+'',r''+newWasRoot+'')

    #endif

    suffix = "process-server/"
    document = configPath + suffix + "config/system/00Static.xml"
    if AdminConfig.existsDocument(document):
      configPath = configPath + suffix
      updateConfigFile(configPath, "config/system/99Sharepoint.xml",r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/system/99Local.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/system/80EventManager.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/HTTPEventListener.properties", r''+oldHostName+'',r''+newHostName+'')

      #update profile root
      updateConfigFile(configPath, "config/system/99Local.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')
      updateConfigFile(configPath, "config/log4j.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')

      #update install root
      updateConfigFile(configPath, "config/system/99Local.xml", r''+oldWasRoot+'',r''+newWasRoot+'')
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldWasRoot+'',r''+newWasRoot+'')
      updateConfigFile(configPath, "config/log4j.xml", r''+oldWasRoot+'',r''+newWasRoot+'')
    #endif

    suffix = "performance-data-warehouse/"
    document = configPath + suffix + "config/system/00Static.xml"
    if AdminConfig.existsDocument(document):
      configPath = configPath + suffix
      updateConfigFile(configPath, "config/system/50AppServer.xml", r''+oldHostName+'',r''+newHostName+'')
      updateConfigFile(configPath, "config/log4j.xml", r''+oldProfileRoot+'',r''+newProfileRoot+'')
      updateConfigFile(configPath, "config/log4j.xml", r''+oldWasRoot+'',r''+newWasRoot+'')
    #endif
  #endfor
#enddef

#------------------------------------------------------------------------------
# This function updates the profile-level configuration for all nodes
# affected by this invocation (may include co-located managed nodes)
#------------------------------------------------------------------------------
def updateProfileConfiguration():
  wasHome = System.getProperty("was.install.root")
  localNodeName = System.getProperty("local.node")
  nodeNames = getNodeNames()
  for nodeName in nodeNames:
    profileRoot = AdminTask.showVariables(["-variableName", "USER_INSTALL_ROOT", "-scope", "Cell=" + cellName + ",Node=" + nodeName])
    profileTempDir = os.path.join(profileRoot, "temp") # AdminTask.showVariables(["-variableName", "WAS_TEMP_DIR", "-scope", "Cell=" + cellName + ",Node=" + nodeName])?
    profile = ProfileHelper.getProfileByLocation(wasHome, profileRoot)
    for template in profile.getTemplates():
      if template.endswith(".procsvr") or template.endswith(".procctr"):
        log("info", "updateProfileConfiguration", "Need to update the following Process Server profile location: " + profileRoot)
        clearIEHSCaches(profileTempDir)
        updateServerPolicyForBPM(profileRoot)
        addedMetadata = addProcessServerConfigMetadata(profileRoot)
        if nodeName == localNodeName:
          localNodeType = profile.getNodeType()
          if localNodeType != ProfileHelper.NODE_TYPE_UNMANAGED and localNodeType != ProfileHelper.NODE_TYPE_DEPLOYMENT_MANAGER:
            log("info", "updateProfileConfiguration", "Profile at " + profileRoot + " is not a Process Server stand-alone or deployment manager profile, skipping cell-level updates.")
          else:
            log("info", "updateProfileConfiguration", "Updating cell-level configuration of " + localNodeType + " profile at " + profileRoot)
            createDefaultFederationDomain(addedMetadata)
            #cell level update
            AdminTask.upgradeBPMConfiguration()
            log("info", "updateProfileConfiguration", "Done updating cell-level configuration of profile at " + profileRoot)
          #endif
        #endif
        break
      #endif
    #endfor
  #endfor
#enddef

#------------------------------------------------------------------------------
# This function clears the IEHS cache
#------------------------------------------------------------------------------
def clearIEHSCaches(profileTempDir):
  def deleteIEHSCacheDirs(args_unused, dirname, names):
    for name in names:
      if name.startswith("IBM_BPM_Help_"):
        cacheDir = os.path.join(dirname, name)
        log("info", "clearIEHSCaches", "Deleting IEHS cache " + cacheDir)
        shutil.rmtree(cacheDir, 1)
      #endif
    #endfor
  #enddef
  log("info", "clearIEHSCaches", "Checking " + profileTempDir + " for IEHS caches")
  os.path.walk(profileTempDir, deleteIEHSCacheDirs, None)
#enddef

#------------------------------------------------------------------------------
# This function adds metadata_procsvr.xml at the config repository
#------------------------------------------------------------------------------
def addProcessServerConfigMetadata(profileRoot):
  metadataFile = os.path.join(profileRoot, "config/.repository/metadata_procsvr.xml")
  # Can't use AdminConfig.createDocument() which is restricted to the local profile, this would not work for co-located managed nodes
  if os.path.exists(metadataFile):
    log("info", "addProcessServerConfigMetadata", "  Process Server Metadata file " + metadataFile + " already exists.")
    return 0
  #endif
  log("info", "addProcessServerConfigMetadata", "  Creating Process Server Metadata file " + metadataFile + ".")
  wasHome = System.getProperty("was.install.root")
  shutil.copy(os.path.join(wasHome, "BPM/base/profile/documents/config/.repository/metadata_procsvr.xml"), metadataFile)
  log("info", "addProcessServerConfigMetadata", "  Created Process Server Metadata Configuration file.")
  return 1
#enddef

#------------------------------------------------------------------------------
# This function adds metadata_procsvr.xml at the config repository
#------------------------------------------------------------------------------
def createDefaultFederationDomain(justAddedMetadata):
  if not "BPMApiDomain" in AdminUtilities.convertToList(AdminConfig.types()):
    log("info", "createDefaultFederationDomain", "  BPMApiDomain WCCM extension not found. Skipping creation of default API federation domain.")
  else:
    cellFedConfig = AdminConfig.getid("/Cell:" + cellName + "/BPMFederationConfiguration:/")
    domainNames = []
    if cellFedConfig:
      # AdminConfig.list("BPMApiDomain", cell) would list all API federation domains in the cell (recursively searches all scopes)
      domainNames = [AdminConfig.showAttribute(domain, "name") for domain in AdminUtilities.convertToList(AdminConfig.list("BPMApiDomain", cellFedConfig))]
    #endif
    if "default" in domainNames:
      log("info", "createDefaultFederationDomain", "  Default API federation domain already exists.")
    else:
      if AdminTask.help("-commands", "createBPMApiFederationDomain").startswith("WASX8024I:"):
        # WASX8024I: Cannot find the command or command group pattern: createBPMApiFederationDomain
        log("info", "createDefaultFederationDomain", "  Assuming a BPM Express installation, skipping creation of default API federation domain.")
      elif justAddedMetadata:
        # need to restart wsadmin JVM before newly copied metadata*.xml files are available
        log("info", "createDefaultFederationDomain", "  BPMApiDomain WCCM extension just added, re-run BPMProfileUpgrade.ant to create the default API federation domain.")
      else:
        log("info", "createDefaultFederationDomain", "  Creating default API federation domain...")
        AdminTask.createBPMApiFederationDomain(["-name", "default", "-addTarget", [[cellName, "", "", ""]]])
      #endif
    #endif
  #endif
#enddef

#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
log("info", "main", __file__ + " started...")

# this will only be set during migration, but not during upgrade
migrationFrom = System.getProperty("bpm.upgrade.previous.version")

log("info", "main", "Updating profile at " + System.getProperty("user.install.root"))
cellName = System.getProperty("local.cell")
hardwareMove = System.getProperty("bpm.upgrade.configuration.copy")
backupDir = os.path.join(System.getProperty("user.install.root"), "configBackup." + time.strftime("%Y-%m-%d_%H-%M", time.localtime()))

if globals().has_key("framework"): # script called from upgrade framework

  clusterName = System.getProperty("bpm.upgrade.cluster.name")
  if not clusterName:
    if hardwareMove == "true":
      log("info", "main", "Performing a hardware move")
      changeHardware()
    #endif
    updateProfileConfiguration()
  #endif

  for appPrefix in apps.keys():
    # Use the framework's "updateDeploymentTargetsOf" function to invoke
    # the above callable (function) for each deployment target where an
    # application whose name starts with 'appPrefix' is deployed.
    appNames = getAppNames('WebSphere:cell=' + cellName, appPrefix)
    updateDeploymentTargetsOf(appNames, updateConfiguration)
  #endfor

else: # script called from command line

  nodeName = None
  serverName = None
  clusterName = None

  if len(sys.argv) == 1:
    clusterName = sys.argv[0]
    scopeObjectName = 'WebSphere:cell=' + cellName + ',cluster=' + clusterName
  elif len(sys.argv) == 2:
    nodeName = sys.argv[0]
    serverName= sys.argv[1]
    scopeObjectName = 'WebSphere:cell=' + cellName + ',node=' + nodeName + ',server=' + serverName
  else:
    log("info", "main", "Usage: wsadmin -conntype none -f " + __file__ + " ( ( <nodeName> <serverName> ) | <clusterName> )")
    log("info", "main", "       The name of the local node is '" + System.getProperty("local.node") + "'")
    sys.exit(-1)
  #endif

  configPath = getScopePath(cellName, nodeName, serverName, clusterName)

  if not AdminConfig.getid(configPath):
    log("error", "main", "Deployment target " + configPath + " does not exist, exiting")
    sys.exit(-1)

  if not clusterName:
    updateProfileConfiguration()

  for appPrefix in apps.keys():
    appNames = getAppNames(scopeObjectName, appPrefix)
    if not appNames:
      log("info", "main", appPrefix + "* not configured on " + configPath + ", skipping")
    for appName in appNames:
      updateConfiguration(appName, nodeName, serverName, clusterName)
  #endfor

#endif

changes = AdminConfig.queryChanges()
if changes.startswith("WASX7241I:"):
  # WASX7241I: There are no unsaved changes in this workspace.
  log("info", "main", "Nothing changed.")
else:
  log("info", "main", "Saving changes, this may take a while...")
  AdminConfig.save()
#endif

log("info", "main", __file__ + " finished.")
